@@ -6,8 +6,12 @@
  * Authors: Iain Holmes  <iain@ximian.com>
  */
 
+#include <string.h>
 #include <libgnome/gnome-util.h>
 #include <libgnome/gnome-i18n.h>
+
+#include <libxml/tree.h>
+
 #include "gnome-cd.h"
 #include "cdrom.h"
 #include "display.h"
@@ -33,19 +37,24 @@ typedef struct _CDImage {
 	GdkRectangle rect;
 } CDImage;
 
-struct _CDDisplayPrivate {
-	GdkColor red;
-	GdkColor blue;
+typedef struct _CDDisplayTheme {
+	gboolean need_resize;
+	
+	CDImage *corners[4];
+	CDImage *straights[4];
+	CDImage *middle;
 
+	CDImage *cd, *track, *loop, *once;
+} CDDisplayTheme;
+
+struct _CDDisplayPrivate {
 	GnomeCDText *layout[CD_DISPLAY_END];
 	int max_width;
 	int height;
 	int need_height;
 
-	CDImage *corners[4];
-	CDImage *straights[4];
-	CDImage *middle;
-
+	CDDisplayTheme *theme;
+	/* These need to be removed */
 	CDImage *cd, *track, *loop, *once;
 
 	GnomeCDRomMode playmode, loopmode;
@@ -87,6 +96,47 @@ scale_image (CDImage *image)
 						 GDK_INTERP_BILINEAR);
 }
 
+static void
+cd_display_resize_images (CDDisplay *display,
+			  GtkAllocation *allocation)
+{
+	CDDisplayPrivate *priv;
+	CDDisplayTheme *theme;
+
+	priv = display->priv;
+	theme = priv->theme;
+
+	if (GTK_WIDGET_REALIZED (GTK_WIDGET (display))) {
+		if (allocation == NULL) {
+			allocation = &(GTK_WIDGET (display)->allocation);
+		}
+	} else {
+		/* Don't need to scale yet */
+		return;
+	}
+
+	theme->need_resize = FALSE;
+	theme->corners[TOPRIGHT]->rect.x = allocation->width - 16;
+	theme->corners[BOTTOMLEFT]->rect.y = allocation->height - 16;
+	theme->corners[BOTTOMRIGHT]->rect.x = allocation->width - 16;
+	theme->corners[BOTTOMRIGHT]->rect.y = allocation->height - 16;
+
+	theme->straights[BOTTOM]->rect.y = allocation->height - 16;
+	theme->straights[RIGHT]->rect.x = allocation->width - 16;
+
+	theme->straights[TOP]->rect.width = allocation->width - 32;
+	theme->straights[BOTTOM]->rect.width = allocation->width - 32;
+	scale_image (theme->straights[TOP]);
+	scale_image (theme->straights[BOTTOM]);
+
+	theme->straights[LEFT]->rect.height = allocation->height - 32;
+	theme->straights[RIGHT]->rect.height = allocation->height - 32;
+
+	theme->middle->rect.width = allocation->width - 32;
+	theme->middle->rect.height = allocation->height - 32;
+	scale_image (theme->middle);
+}
+	
 static void
 size_allocate (GtkWidget *drawing_area,
 	       GtkAllocation *allocation)
@@ -95,7 +145,7 @@ size_allocate (GtkWidget *drawing_area,
 	CDDisplayPrivate *priv;
 	PangoContext *context;
 	PangoDirection base_dir;
-	int i, mod;
+	int i;
 	
 	disp = CD_DISPLAY (drawing_area);
 	priv = disp->priv;
@@ -114,26 +164,7 @@ size_allocate (GtkWidget *drawing_area,
 	}
 
 	/* Resize and position pixbufs */
-	priv->corners[TOPRIGHT]->rect.x = allocation->width - 16;
-	priv->corners[BOTTOMLEFT]->rect.y = allocation->height - 16;
-	priv->corners[BOTTOMRIGHT]->rect.x = allocation->width - 16;
-	priv->corners[BOTTOMRIGHT]->rect.y = allocation->height - 16;
-
-	priv->straights[BOTTOM]->rect.y = allocation->height - 16;
-	priv->straights[RIGHT]->rect.x = allocation->width - 16;
-
-	priv->straights[TOP]->rect.width = allocation->width - 32;
-	priv->straights[BOTTOM]->rect.width = allocation->width - 32;
-	scale_image (priv->straights[TOP]);
-	scale_image (priv->straights[BOTTOM]);
-
-	priv->straights[LEFT]->rect.height = allocation->height - 32;
-	priv->straights[RIGHT]->rect.height = allocation->height - 32;
-
-	priv->middle->rect.width = allocation->width - 32;
-	priv->middle->rect.height = allocation->height - 32;
-	scale_image (priv->middle);
-	
+	cd_display_resize_images (disp, allocation);
 	GTK_WIDGET_CLASS (parent_class)->size_allocate (drawing_area, allocation);
 }
 
@@ -143,7 +174,7 @@ size_request (GtkWidget *widget,
 {
 	CDDisplay *disp;
 	CDDisplayPrivate *priv;
-	int i, height = 0, width = 0, mod;
+	int i, height = 0, width = 0;
 	
 	disp = CD_DISPLAY (widget);
 	priv = disp->priv;
@@ -198,27 +229,36 @@ expose_event (GtkWidget *drawing_area,
 {
 	CDDisplay *disp;
 	CDDisplayPrivate *priv;
+	CDDisplayTheme *theme;
 	PangoContext *context;
 	PangoDirection base_dir;
 	GdkRectangle *area;
+	GtkStateType state;
 	int height = 0;
 	int i;
 
 	disp = CD_DISPLAY (drawing_area);
 	priv = disp->priv;
+	theme = priv->theme;
 
+	g_assert (theme != NULL);
+	
 	area = &event->area;
 
+	state = GTK_WIDGET_STATE (drawing_area);
+	if (theme->need_resize == TRUE) {
+		cd_display_resize_images (disp, NULL);
+	}
 	/* Check corners and draw them */
 	for (i = 0; i < 4; i++) {
 		GdkRectangle inter;
 		
-		if (gdk_rectangle_intersect (area, &priv->corners[i]->rect, &inter) == TRUE) {
-			draw_pixbuf (priv->corners[i]->pixbuf,
+		if (gdk_rectangle_intersect (area, &theme->corners[i]->rect, &inter) == TRUE) {
+			draw_pixbuf (theme->corners[i]->pixbuf,
 				     drawing_area->window,
 				     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-				     inter.x - priv->corners[i]->rect.x,
-				     inter.y - priv->corners[i]->rect.y,
+				     inter.x - theme->corners[i]->rect.x,
+				     inter.y - theme->corners[i]->rect.y,
 				     inter.x, inter.y,
 				     inter.width, inter.height);
 		}
@@ -227,12 +267,12 @@ expose_event (GtkWidget *drawing_area,
 	for (i = TOP; i <= BOTTOM; i++) {
 		GdkRectangle inter;
 
-		if (gdk_rectangle_intersect (&priv->straights[i]->rect, area, &inter) == TRUE) {
-			draw_pixbuf (priv->straights[i]->scaled,
+		if (gdk_rectangle_intersect (&theme->straights[i]->rect, area, &inter) == TRUE) {
+			draw_pixbuf (theme->straights[i]->scaled,
 				     drawing_area->window,
 				     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-				     inter.x - priv->straights[i]->rect.x,
-				     inter.y - priv->straights[i]->rect.y,
+				     inter.x - theme->straights[i]->rect.x,
+				     inter.y - theme->straights[i]->rect.y,
 				     inter.x, inter.y,
 				     inter.width, inter.height);
 		}
@@ -240,17 +280,17 @@ expose_event (GtkWidget *drawing_area,
 
 	for (i = LEFT; i <= RIGHT; i++) {
 		GdkRectangle inter;
-		if (gdk_rectangle_intersect (area, &priv->straights[i]->rect, &inter) == TRUE) {
+		if (gdk_rectangle_intersect (area, &theme->straights[i]->rect, &inter) == TRUE) {
 			int repeats, extra_s, extra_end, d, j;
 
 			d = inter.y / 16;
 			extra_s = inter.y - (d * 16);
 			
 			if (extra_s > 0) {
-				draw_pixbuf (priv->straights[i]->pixbuf,
+				draw_pixbuf (theme->straights[i]->pixbuf,
 					     drawing_area->window,
 					     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-					     inter.x - priv->straights[i]->rect.x,
+					     inter.x - theme->straights[i]->rect.x,
 					     16 - extra_s,
 					     inter.x, inter.y,
 					     inter.width, extra_s);
@@ -260,10 +300,10 @@ expose_event (GtkWidget *drawing_area,
 			extra_end = (inter.height - extra_s) % 16;
 			
 			for (j = 0; j < repeats; j++) {
-				draw_pixbuf (priv->straights[i]->pixbuf,
+				draw_pixbuf (theme->straights[i]->pixbuf,
 					     drawing_area->window,
 					     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-					     inter.x - priv->straights[i]->rect.x,
+					     inter.x - theme->straights[i]->rect.x,
 					     0,
 					     inter.x,
 					     inter.y + (j * 16) + extra_s,
@@ -271,10 +311,10 @@ expose_event (GtkWidget *drawing_area,
 			}
 
 			if (extra_end > 0) {
-				draw_pixbuf (priv->straights[i]->pixbuf,
+				draw_pixbuf (theme->straights[i]->pixbuf,
 					     drawing_area->window,
 					     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-					     inter.x - priv->straights[i]->rect.x,
+					     inter.x - theme->straights[i]->rect.x,
 					     0,
 					     inter.x,
 					     inter.y + (repeats * 16) + extra_s,
@@ -286,17 +326,17 @@ expose_event (GtkWidget *drawing_area,
 	/* Do the middle - combination of the above */
 	{
 		GdkRectangle inter;
-		if (gdk_rectangle_intersect (area, &priv->middle->rect, &inter) == TRUE) {
+		if (gdk_rectangle_intersect (area, &theme->middle->rect, &inter) == TRUE) {
 			int repeats, extra_s, extra_end, j, d;
 
 			d = inter.y / 16;
 			extra_s = inter.y - (d * 16);
 			
 			if (extra_s > 0) {
-				draw_pixbuf (priv->middle->scaled,
+				draw_pixbuf (theme->middle->scaled,
 					     drawing_area->window,
 					     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-					     inter.x - priv->middle->rect.x,
+					     inter.x - theme->middle->rect.x,
 					     16 - extra_s,
 					     inter.x, inter.y,
 					     inter.width, extra_s);
@@ -306,10 +346,10 @@ expose_event (GtkWidget *drawing_area,
 			extra_end = (inter.height - extra_s) % 16;
 			
 			for (j = 0; j < repeats; j++) {
-				draw_pixbuf (priv->middle->scaled,
+				draw_pixbuf (theme->middle->scaled,
 					     drawing_area->window,
 					     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-					     inter.x - priv->middle->rect.x,
+					     inter.x - theme->middle->rect.x,
 					     0,
 					     inter.x,
 					     inter.y + (j * 16) + extra_s,
@@ -317,10 +357,10 @@ expose_event (GtkWidget *drawing_area,
 			}
 
 			if (extra_end > 0) {
-				draw_pixbuf (priv->middle->scaled,
+				draw_pixbuf (theme->middle->scaled,
 					     drawing_area->window,
 					     drawing_area->style->bg_gc[GTK_STATE_NORMAL],
-					     inter.x - priv->middle->rect.x,
+					     inter.x - theme->middle->rect.x,
 					     0,
 					     inter.x,
 					     inter.y + (repeats * 16) + extra_s,
@@ -369,13 +409,18 @@ expose_event (GtkWidget *drawing_area,
 						    base_dir == PANGO_DIRECTION_LTR ? PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT);
 
 			gdk_draw_layout_with_colors (drawing_area->window,
-						     drawing_area->style->black_gc,
+						     drawing_area->style->text_gc[state],
 						     X_OFFSET, 
 						     Y_OFFSET + height,
 						     priv->layout[i]->layout,
-						     &priv->red, NULL);
+						     NULL, NULL);
+		}
+
+		if (i == CD_DISPLAY_LINE_INFO) {
+			height += 16;
+		} else {
+			height += priv->layout[i]->height;
 		}
-		height += priv->layout[i]->height;
 	}
 
 	return TRUE;
@@ -412,9 +457,10 @@ static void
 realize (GtkWidget *widget)
 {
 	CDDisplay *disp;
-	GdkColormap *cmap;
 
 	disp = CD_DISPLAY (widget);
+#if 0
+	GdkColormap *cmap;
 
 	cmap = gtk_widget_get_colormap (widget);
 	disp->priv->red.red = 65535;
@@ -428,9 +474,10 @@ realize (GtkWidget *widget)
 	disp->priv->blue.blue = 65535;
 	disp->priv->blue.pixel = 0;
 	gdk_color_alloc (cmap, &disp->priv->blue);
-
+#endif
 	gtk_widget_add_events (widget, GDK_BUTTON_PRESS_MASK |
 			       GDK_KEY_PRESS_MASK);
+
 	GTK_WIDGET_CLASS (parent_class)->realize (widget);
 }
 
@@ -526,25 +573,31 @@ cd_image_new (const char *filename,
 	CDImage *image;
 	char *fullname;
 	
-	image = g_new (CDImage, 1);
+	image = g_new0 (CDImage, 1);
 
-	fullname = gnome_pixmap_file (filename);
-	if (fullname == NULL) {
-		g_warning ("Error loading %s", filename);
-		g_free (image);
-		return NULL;
+	if (filename[0] != '/') {
+		fullname = gnome_program_locate_file (NULL,
+			   GNOME_FILE_DOMAIN_PIXMAP, filename, TRUE, NULL);
+		if (fullname == NULL) {
+			g_warning ("Error loading %s", filename);
+			g_free (image);
+			return NULL;
+		}
+		
+		image->pixbuf = gdk_pixbuf_new_from_file (fullname, NULL);
+		g_free (fullname);
+	} else {
+		image->pixbuf = gdk_pixbuf_new_from_file (filename, NULL);
 	}
 	
-	image->pixbuf = gdk_pixbuf_new_from_file (fullname, NULL);
-	g_free (fullname);
-	
 	if (image->pixbuf == NULL) {
 		g_warning ("Error loading %s", filename);
 		g_free (image);
 		return NULL;
 	}
 
-	image->scaled = NULL;
+	image->scaled = image->pixbuf;
+	g_object_ref (image->scaled);
 	image->rect.x = x;
 	image->rect.y = y;
 	image->rect.width = gdk_pixbuf_get_width (image->pixbuf);
@@ -567,10 +620,10 @@ init (CDDisplay *disp)
 	for (i = 0; i < CD_DISPLAY_END; i++) {
 		PangoRectangle rect;
 		
-		priv->layout[i] = g_new (GnomeCDText, 1);
+		priv->layout[i] = g_new0 (GnomeCDText, 1);
 		
-		priv->layout[i]->text = g_strdup (default_text[i]);
-		priv->layout[i]->length = strlen (default_text[i]);
+		priv->layout[i]->text = g_strdup (_(default_text[i]));
+		priv->layout[i]->length = strlen (_(default_text[i]));
 		priv->layout[i]->layout = gtk_widget_create_pango_layout (GTK_WIDGET (disp), priv->layout[i]->text);
 		pango_layout_set_text (priv->layout[i]->layout,
 				       priv->layout[i]->text,
@@ -590,19 +643,6 @@ init (CDDisplay *disp)
 	priv->playmode = GNOME_CDROM_WHOLE_CD;
 	priv->loopmode = GNOME_CDROM_PLAY_ONCE;
 	
-	/* Load pixbufs */
-	priv->corners[TOPLEFT] = cd_image_new ("gnome-cd/lcd-theme/top-left.png", 0, 0);
-	priv->corners[TOPRIGHT] = cd_image_new ("gnome-cd/lcd-theme/top-right.png", 48, 0);
-	priv->corners[BOTTOMLEFT] = cd_image_new ("gnome-cd/lcd-theme/bottom-left.png", 0, 48);
-	priv->corners[BOTTOMRIGHT] = cd_image_new ("gnome-cd/lcd-theme/bottom-right.png", 48, 48);
-
-	priv->straights[LEFT] = cd_image_new ("gnome-cd/lcd-theme/middle-left.png", 0, 16);
-	priv->straights[RIGHT] = cd_image_new ("gnome-cd/lcd-theme/middle-right.png", 48, 16);
-	priv->straights[TOP] = cd_image_new ("gnome-cd/lcd-theme/top.png", 16, 0);
-	priv->straights[BOTTOM] = cd_image_new ("gnome-cd/lcd-theme/bottom.png", 16, 48);
-
-	priv->middle = cd_image_new ("gnome-cd/lcd-theme/middle.png", 16, 16);
-
 	/* Don't know where these are to be placed yet */
 	priv->track = cd_image_new ("gnome-cd/track.png", 16, 0);
 	priv->cd = cd_image_new ("gnome-cd/disc.png", 16, 0);
@@ -664,8 +704,13 @@ cd_display_set_line (CDDisplay *disp,
 	g_return_if_fail (new_str != NULL);
 
 	priv = disp->priv;
-	
+
 	text = priv->layout[line];
+	if (strcmp (new_str, text->text) == 0) {
+		/* Same, do nothing */
+		return;
+	}
+
 	height = priv->height - text->height;
 	
 	g_free (text->text);
@@ -712,3 +757,94 @@ cd_display_clear (CDDisplay *disp)
 	priv->max_width = max_width;
 	gtk_widget_queue_resize (GTK_WIDGET (disp));
 }
+
+static inline char *
+make_fullname (const char *theme_name,
+	       const char *name)
+{
+	char *image;
+
+	image = g_build_filename (THEME_DIR, theme_name, name, NULL);
+	
+	return image;
+}
+
+void
+cd_display_parse_theme (CDDisplay *disp,
+			GCDTheme *cd_theme,
+			xmlDocPtr doc,
+			xmlNodePtr cur)
+{
+	CDDisplayPrivate *priv;
+	CDDisplayTheme *theme;
+
+	priv = disp->priv;
+	/* Should probably destroy the old theme here */
+
+	priv->theme = g_new0 (CDDisplayTheme, 1);
+	theme = priv->theme;
+
+	theme->need_resize = TRUE;
+	while (cur != NULL) {
+		if (xmlStrcmp (cur->name, (const xmlChar *) "image") == 0) {
+			xmlChar *location;
+
+			location = xmlGetProp (cur, (const xmlChar *) "location");
+			if (location != NULL) {
+				char *file, *full;
+
+				file = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
+				full = make_fullname (cd_theme->name, file);
+				
+				if (xmlStrcmp (location, "top-left") == 0) {
+					theme->corners[TOPLEFT] = cd_image_new (full, 0, 0);
+				} else if (xmlStrcmp (location, "top-right") == 0) {
+					theme->corners[TOPRIGHT] = cd_image_new (full, 48, 0);
+				} else if (xmlStrcmp (location, "bottom-right") == 0) {
+					theme->corners[BOTTOMRIGHT] = cd_image_new (full, 48, 32);
+				} else if (xmlStrcmp (location, "bottom-left") == 0) {
+					theme->corners[BOTTOMLEFT] = cd_image_new (full, 0, 32);
+				} else if (xmlStrcmp (location, "left") == 0) {
+					theme->straights[LEFT] = cd_image_new (full, 0, 16);
+				} else if (xmlStrcmp (location, "right") == 0) {
+					theme->straights[RIGHT] = cd_image_new (full, 48, 16);
+				} else if (xmlStrcmp (location, "top") == 0) {
+					theme->straights[TOP] = cd_image_new (full, 16, 0);
+				} else if (xmlStrcmp (location, "bottom") == 0) {
+					theme->straights[BOTTOM] = cd_image_new (full, 16, 32);
+				} else if (xmlStrcmp (location, "middle") == 0) {
+					theme->middle = cd_image_new (full, 16, 16);
+				} else {
+					/** Hmmmmm */
+				}
+
+				g_free (full);
+			}
+		}
+
+		cur = cur->next;
+	}
+
+	/* Update sizes */
+	cd_display_resize_images (disp, NULL);
+}
+
+GnomeCDText *
+cd_display_get_layout(CDDisplay* disp,
+		      int i)
+{
+	g_return_val_if_fail (disp != NULL, NULL);
+	g_return_val_if_fail (i < CD_DISPLAY_END, NULL);
+
+	return disp->priv->layout[i];
+}
+
+PangoLayout *
+cd_display_get_pango_layout(CDDisplay *disp,
+			    int i)
+{
+	g_return_val_if_fail (disp != NULL, NULL);
+	g_return_val_if_fail (i < CD_DISPLAY_END, NULL);
+
+	return disp->priv->layout[i]->layout;
+}
